""" This file describe the API to get the state of messages """
import sys
from typing import List

from flask import Flask
from flask_apispec import marshal_with
from marshmallow import Schema, ValidationError, fields, validate, validates_schema
from typing_extensions import TypedDict
from werkzeug.exceptions import HTTPException

import remoulade
from remoulade import get_broker
from remoulade.errors import NoResultBackend, NoStateBackend, RemouladeError
from remoulade.result import Result
from remoulade.results import ResultMissing

from .apispec import add_swagger, validate_schema
from .scheduler import scheduler_bp, scheduler_routes
from .state import messages_bp, messages_routes

app = Flask(__name__)
app.register_blueprint(scheduler_bp)
app.register_blueprint(messages_bp)


class MessageSchema(Schema):
    """
    Class to validate post data in /messages
    """

    actor_name = fields.Str(validate=validate.Length(min=1), required=True)
    args = fields.List(fields.Raw(), allow_none=True)
    kwargs = fields.Dict(allow_none=True)
    options = fields.Dict(allow_none=True)
    delay = fields.Number(validate=validate.Range(min=1), allow_none=True)

    @validates_schema
    def validate_actor_name(self, data, **kwargs):
        actor_name = data.get("actor_name")
        if actor_name not in remoulade.get_broker().actors.keys():
            raise ValidationError(f"No actor named {actor_name} exists")


class ResponseSchema(Schema):
    result = fields.Raw(allow_none=True)
    error = fields.Str(allow_none=True)


class ArgsSchema(Schema):
    name = fields.Str()
    type = fields.Str()
    default = fields.Str(allow_none=True)


class ActorSchema(Schema):
    name = fields.Str()
    queue_name = fields.Str()
    alternative_queues = fields.List(fields.Str())
    priority = fields.Int()
    args = fields.List(fields.Nested(ArgsSchema))


class ActorResponseSchema(Schema):
    result = fields.List(fields.Nested(ActorSchema))


class OptionsResponseSchema(Schema):
    options = fields.List(fields.Str())


@app.route("/messages/cancel/<message_id>", methods=["POST"])
@marshal_with(ResponseSchema)
def cancel_message(message_id):
    broker = get_broker()
    cancel_backend = broker.get_cancel_backend()
    try:
        # If a single message in a composition is canceled, we cancel the whole composition
        state_backend = broker.get_state_backend()
        states_count = state_backend.get_states_count(selected_composition_ids=[message_id])
        state = state_backend.get_state(message_id)
        if states_count == 0 and state is None:
            raise ValidationError(f"This message id or composition id {message_id} does not exist.")
        if state and state.composition_id:
            message_id = state.composition_id
    except NoStateBackend:
        pass
    cancel_backend.cancel([message_id])
    return {"result": "ok"}


@app.route("/messages/requeue/<message_id>", methods=["POST"])
@marshal_with(ResponseSchema)
def requeue_message(message_id):
    broker = get_broker()
    backend = broker.get_state_backend()
    state = backend.get_state(message_id)
    if state is None:
        raise ValidationError(f"No message with id {message_id} exists")
    actor = broker.get_actor(state.actor_name)
    payload = {"args": state.args, "kwargs": state.kwargs}
    pipe_target = state.options.get("pipe_target")
    if pipe_target is None:
        actor.send_with_options(**payload, **state.options)
        return {"result": "ok"}
    else:
        return {"error": "requeue message in a pipeline not supported"}, 400


@app.route("/messages/result/<message_id>")
@marshal_with(ResponseSchema)
def get_results(message_id):
    from ..message import get_encoder

    max_size = 1e4
    try:
        result = Result(message_id=message_id).get()
        encoded_result = get_encoder().encode(result).decode("utf-8")
        size_result = sys.getsizeof(encoded_result)
        if size_result >= max_size:
            encoded_result = f"The result is too big {size_result / 1e6}M"
        return {"result": encoded_result}
    except ResultMissing:
        return {"result": "result is missing"}
    except NoResultBackend:
        return {"result": "no result backend"}
    except (UnicodeDecodeError, TypeError):
        return {"result": "non serializable result"}


@app.route("/messages", methods=["POST"])
@marshal_with(ResponseSchema)
@validate_schema(MessageSchema)
def enqueue_message(**kwargs):
    actor = get_broker().get_actor(kwargs.pop("actor_name"))
    options = kwargs.pop("options") or {}
    actor.send_with_options(**kwargs, **options)
    return {"result": "ok"}


class GroupMessagesT(TypedDict):
    group_id: str
    messages: List[dict]


@app.route("/actors")
@marshal_with(ActorResponseSchema)
def get_actors():
    return {"result": [actor.as_dict() for actor in get_broker().actors.values()]}


@app.route("/options")
@marshal_with(OptionsResponseSchema)
def get_options():
    broker = get_broker()
    return {"options": list(broker.actor_options)}


@app.errorhandler(RemouladeError)
def remoulade_exception(e):
    return {"error": str(e)}, 500


@app.errorhandler(HTTPException)
def http_exception(e):
    return {"error": str(e)}, e.code


@app.errorhandler(ValidationError)
def validation_error(e):
    return {"error": e.normalized_messages()}, 400


routes = [cancel_message, requeue_message, get_results, enqueue_message, get_actors, get_options]

add_swagger(app, {"main": routes, "scheduler": scheduler_routes, "messages": messages_routes})



import datetime
import sys
from collections import namedtuple
from enum import Enum
from typing import Dict, List, Optional

from dateutil.parser import parse

from ..encoder import Encoder
from .errors import InvalidStateError


class StateStatusesEnum(Enum):
    """Contains the possible statuses that a message can have"""

    Started = "Started"  # a `Message` that has not been processed
    Pending = "Pending"  # a `Message` that has been enqueued
    Skipped = "Skipped"  # a `Message` that has been skipped
    Canceled = "Canceled"  # a `Message` that has been canceled
    Failure = "Failure"  # a `Message` that has been processed and raised an Exception
    Success = "Success"  # a `Message` that has been processed and does not raise an Exception


#: A type alias representing states in the database
class State(
    namedtuple(
        "State",
        (
            "message_id",
            "status",
            "actor_name",
            "args",
            "kwargs",
            "options",
            "priority",
            "progress",
            "enqueued_datetime",
            "started_datetime",
            "end_datetime",
            "queue_name",
            "composition_id",
        ),
    )
):
    """Catalog Class, it storages the state
    Parameters:
        status: The current status of the message state
        args: List of arguments in the state
    """

    def __new__(
        cls,
        message_id,
        status=None,
        *,
        actor_name=None,
        args=None,
        kwargs=None,
        options=None,
        priority=None,
        progress=None,
        enqueued_datetime=None,
        started_datetime=None,
        end_datetime=None,
        queue_name=None,
        composition_id=None,
    ):

        if status and status not in list(StateStatusesEnum):
            raise InvalidStateError(f"The {status} State is not defined")
        return super().__new__(
            cls,
            message_id,
            status,
            actor_name,
            args,
            kwargs,
            options,
            priority,
            progress,
            enqueued_datetime,
            started_datetime,
            end_datetime,
            queue_name,
            composition_id,
        )

    def as_dict(self, exclude_keys=(), encode_args=False):
        """Transform a State into a dict, can exclude some keys"""
        as_dict = {
            key: value for (key, value) in self._asdict().items() if value is not None and key not in exclude_keys
        }
        datetime_keys = ["enqueued_datetime", "started_datetime", "end_datetime"]
        for key in datetime_keys:
            if key in as_dict:
                as_dict[key] = as_dict[key].isoformat()
        if self.status:
            as_dict["status"] = self.status.value
        if encode_args:
            from ..message import get_encoder

            for key in (item for item in ["args", "kwargs", "options"] if item in as_dict):
                try:
                    as_dict[key] = get_encoder().encode(as_dict[key]).decode("utf-8")
                except (UnicodeDecodeError, TypeError):
                    as_dict[key] = "encoded_data"
        return as_dict

    @classmethod
    def from_dict(cls, input_dict: Dict) -> "State":
        if "status" in input_dict:
            input_dict["status"] = StateStatusesEnum(input_dict["status"])
        datetime_keys = ["enqueued_datetime", "started_datetime", "end_datetime"]
        for key in datetime_keys:
            if key in input_dict and type(input_dict[key]) == str:
                input_dict[key] = parse(input_dict[key])
        return cls(**input_dict)


class StateBackend:
    """ABC for  state backends.

    Parameters:
      namespace(str): The logical namespace under which the data
        should be stored.
      encoder(Encoder): The encoder to use when storing and retrieving
        result data.  Defaults to :class:`.JSONEncoder`.
      max_size(int): Maximum size of arguments allow to storage
        in the database, default 2MB
    """

    namespace = "remoulade-state*"

    def __init__(self, *, namespace: str = "remoulade-state", encoder: Encoder = None, max_size=2e6):
        from ..message import get_encoder

        self.namespace = namespace
        self.encoder = encoder or get_encoder()
        self.max_size = max_size

    def _build_message_key(self, message_id: str) -> str:  # noqa: F821
        """Given a message id, return its globally-unique key.

        Parameters:
          message_id(str)

        Returns:
          str
        """
        return f"{self.namespace}:{message_id}"

    def get_state(self, message_id: str) -> State:
        """Get the state with a message_id from the backend.

        Parameters:
            message_id(str)

        """
        raise NotImplementedError(f"{type(self).__name__} does not implement get_state")

    def set_state(self, state: State, ttl: int = 3600) -> None:
        """Save a message in the backend if it does not exist,
            otherwise update it.

        Parameters:
            state(State)
            ttl(seconds): The time to keep that state in the backend
             default is one hour(3600 seconds)
        """
        raise NotImplementedError(f"{type(self).__name__} does not implement set_state")

    def get_states(
        self,
        *,
        size: Optional[int] = None,
        offset: int = 0,
        selected_actors: Optional[List[str]] = None,
        selected_statuses: Optional[List[str]] = None,
        selected_message_ids: Optional[List[str]] = None,
        selected_composition_ids: Optional[List[str]] = None,
        start_datetime: Optional[datetime.datetime] = None,
        end_datetime: Optional[datetime.datetime] = None,
        sort_column: Optional[str] = None,
        sort_direction: Optional[str] = None,
    ) -> List[State]:
        """Return all the states in the backend"""
        raise NotImplementedError(f"{type(self).__name__} does not implement get_all_messages")

    def get_states_count(
        self,
        *,
        selected_actors: Optional[List[str]] = None,
        selected_statuses: Optional[List[str]] = None,
        selected_messages_ids: Optional[List[str]] = None,
        selected_composition_ids: Optional[List[str]] = None,
        start_datetime: Optional[datetime.datetime] = None,
        end_datetime: Optional[datetime.datetime] = None,
        **kwargs,
    ) -> int:
        raise NotImplementedError(f"{type(self).__name__} does not implement get_states_count")

    def _encode_dict(self, data):
        """Return the (keys, values) of a dictionary encoded"""
        encoded_data = {}
        for (key, value) in data.items():
            encoded_value = self.encoder.encode(value)
            if sys.getsizeof(encoded_value) <= self.max_size:
                encoded_data[self.encoder.encode(key)] = self.encoder.encode(value)
        return encoded_data

    def _decode_dict(self, data):
        """Return the (keys, values) of a dictionary decoded"""
        decoded_data = {}
        for (key, value) in data.items():
            decoded_data[self.encoder.decode(key)] = self.encoder.decode(value)
        return decoded_data

    def clean(self, max_age: int = None, not_started: bool = False):
        """Deletes states.

        Parameters:
            max_age(int): When set, only delete states that have finished more than max_age minutes ago.
            not_started(bool): When set to True, only delete states from messages that have not yet started."""
        raise NotImplementedError(f"{type(self).__name__} does not implement clean")



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import json
import pickle
import warnings
from typing import Any, Dict, Optional, get_type_hints

try:
    from pydantic import BaseConfig, BaseModel, ValidationError, create_model
    from pydantic.error_wrappers import ErrorWrapper
    from pydantic.fields import ModelField
    from simplejson.decoder import JSONDecoder
    from simplejson.encoder import JSONEncoder as _JSONEncoder
except ImportError:  # pragma: no cover
    warnings.warn(
        "Pydantic and simplejson are not available.  Run `pip install remoulade[pydantic]`",
        ImportWarning,
    )


from .errors import ActorNotFound

#: Represents the contents of a Message object as a dict.
MessageData = Dict[str, Any]


class Encoder(abc.ABC):
    """Base class for message encoders."""

    @abc.abstractmethod
    def encode(self, data: MessageData) -> bytes:  # pragma: no cover
        """Convert message metadata into a bytestring."""
        raise NotImplementedError

    @abc.abstractmethod
    def decode(self, data: bytes) -> MessageData:  # pragma: no cover
        """Convert a bytestring into message metadata."""
        raise NotImplementedError


class JSONEncoder(Encoder):
    """Encodes messages as JSON.  This is the default encoder."""

    def encode(self, data: MessageData) -> bytes:
        return json.dumps(data, separators=(",", ":")).encode("utf-8")

    def decode(self, data: bytes) -> MessageData:
        return json.loads(data.decode("utf-8"))


class PickleEncoder(Encoder):
    """Pickles messages.

    Warning:
      This encoder is not secure against maliciously-constructed data.
      Use it at your own risk.
    """

    encode = pickle.dumps  # type: ignore
    decode = pickle.loads  # type: ignore


class PydanticEncoder(Encoder):
    """PydanticEncoder remoulade encoder working with Pydantic schemas (install remoulade[pydantic] extra dependency)
    With this encoder you must use only Pydantic schema as inputs/outputs of the actors and type them explicitly.

    class MyActorInputSchema(BaseModel):
        ...

    class MyActorOutputSchema(BaseModel):
        ...

    @remoulade.actor()
    def my_actor(input_1: MyActorInputSchema, input_2: MyActorInputSchema | None = None) -> MyActorOutputSchema:
        ...
        return MyActorOutputSchema()
    """

    def __init__(self, fallback_encoder: Optional[Encoder] = None):
        self.fallback_encoder = fallback_encoder
        self.json_encoder = _JSONEncoder(default=self.default)
        self.json_decoder = JSONDecoder()

    @staticmethod
    def default(o):
        if isinstance(o, BaseModel):
            # keep dict otherwise it will be serialized as a string (see Pydantic .json())
            return json.loads(o.json())
        raise TypeError("Object of type %s is not JSON serializable" % o.__class__.__name__)

    def encode(self, data: MessageData) -> bytes:
        try:
            return self.json_encoder.encode(data).encode("utf-8")
        except Exception as e:
            if self.fallback_encoder is not None:
                return self.fallback_encoder.encode(data)
            else:
                raise e

    def decode(self, data: bytes) -> MessageData:
        from remoulade import get_broker

        try:
            raw_message = self.json_decoder.decode(data.decode("utf-8"))  # type: ignore
            actor_name = raw_message["actor_name"]
            actor_fn = get_broker().get_actor(actor_name).fn

            # Retrieve the Pydantic schemas from typing
            schemas_by_param_name = {}
            for param_name, type_hint in get_type_hints(actor_fn).items():
                schemas_by_param_name[param_name] = ModelField(
                    type_=type_hint,
                    name=f"{param_name}_schema",
                    class_validators=None,
                    model_config=BaseConfig,
                )

            # Override message_data with Pydantic schema when it matches
            parsed_message: Dict[str, Any] = {}
            for key, values in raw_message.items():
                if key == "kwargs":
                    assert isinstance(values, dict)
                    parsed_message[key] = {
                        param_name: self.get_parsed_value(raw_value, schemas_by_param_name[param_name])
                        for param_name, raw_value in values.items()
                    }
                elif key == "args":
                    assert isinstance(values, list)
                    schemas = list(schemas_by_param_name.values())
                    parsed_message[key] = [
                        self.get_parsed_value(raw_value, schemas[order]) for order, raw_value in enumerate(values)
                    ]
                elif key == "result":
                    parsed_value = self.get_parsed_value(values, schemas_by_param_name["return"])
                    try:
                        _parsed_value = list(parsed_value)
                    except TypeError:
                        _parsed_value = parsed_value
                    parsed_message[key] = _parsed_value
                else:
                    parsed_message[key] = values

            return parsed_message
        except Exception as e:
            if self.fallback_encoder is not None:
                return self.fallback_encoder.decode(data)
            else:
                raise e

    @staticmethod
    def get_parsed_value(raw_value: Any, schema: "ModelField"):
        errors = []
        parsed_value, errors_ = schema.validate(raw_value, {}, loc=())
        if isinstance(errors_, ErrorWrapper):
            errors.append(errors_)
        elif isinstance(errors_, list):
            errors.extend(errors_)
        if errors:
            raise ValidationError(errors, create_model("RemouladeEncoder"))
        return parsed_value
